import json
import sys
from copy import deepcopy
from pathlib import Path

import pytest
import yaml

from naapc import ndict

ROOT = Path(__file__).resolve().parents[1]
TEST_SRC_DIR = ROOT / "test"
TEST_ASSET = TEST_SRC_DIR / "assets"
if str(TEST_SRC_DIR) not in sys.path:
    sys.path.append(str(TEST_SRC_DIR))

from utils import get_dict_compare_msg


def test_init():
    with open(TEST_ASSET / "init.json", "r") as f:
        raw = json.load(f)
    with open(TEST_ASSET / "init_flatten.json", "r") as f:
        flatten = json.load(f)
    with open(TEST_ASSET / "init_nested.json", "r") as f:
        nested = json.load(f)

    d = ndict(raw)
    assert d.dict == nested, get_dict_compare_msg(d.dict, nested, indent=4, sort_keys=False)
    assert d.flatten_dict == flatten, get_dict_compare_msg(
        d.flatten_dict, flatten, indent=4, sort_keys=False
    )

    d1 = ndict(d)
    assert d1.dict is d.dict
    assert d1.flatten_dict is d.flatten_dict

    d2 = ndict(flatten)
    assert d2.dict == nested, get_dict_compare_msg(d2.dict, nested, indent=4, sort_keys=False)
    assert d2.flatten_dict == flatten, get_dict_compare_msg(
        d2.flatten_dict, flatten, indent=4, sort_keys=False
    )

    d = ndict()
    assert d.dict == {}
    assert d.flatten_dict == {}
    assert d._d == {}
    assert d._flatten_dict == {}


def test_getitem():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))

    with open(TEST_ASSET / "init_flatten.json", "r") as f:
        flatten = json.load(f)

    for p, v in flatten.items():
        assert d[p] == v

    with open(TEST_ASSET / "init_getitem.json", "r") as f:
        getitem_gt = json.load(f)

    for p, gt in getitem_gt.items():
        v = d[p]
        if isinstance(gt, dict):
            assert isinstance(v, ndict)
            v = v.dict
        assert v == gt

    with pytest.raises(KeyError):
        a = d["not_exist_path"]


def test_delitem():
    with open(TEST_ASSET / "init.json", "r") as f:
        raw = json.load(f)
    with open(TEST_ASSET / "init_del.json", "r") as f:
        del_gt = json.load(f)
    for p, gt in del_gt.items():
        d = ndict(deepcopy(raw))
        del d[p]
        assert d.dict == gt["dict"]
        assert d.flatten_dict == gt["flatten"]

    with open(TEST_ASSET / "init_nested.json", "r") as f:
        nested = json.load(f)
    d = ndict(deepcopy(raw))
    for k in nested.keys():
        del d[k]
    assert d.dict == {}
    assert d.flatten_dict == {}


def test_setitem():
    with open(TEST_ASSET / "init.json", "r") as f:
        raw = json.load(f)
    d = ndict(raw)
    d["not_exist"] = {"something1": {"something2": 1}}
    assert d["not_exist;something1;something2"] == 1
    d["not_exist"] = {}
    assert "not_exist;something1;something2" not in d.flatten_dict

    d = ndict(raw)
    d["not_exist"] = {"something1": {"something2": 1}}
    assert d["not_exist;something1;something2"] == 1
    d["not_exist;something1"] = 1
    assert "not_exist;something1;something2" not in d.flatten_dict

    d = ndict()
    d["something"] = {}
    assert d.dict == {"something": {}}
    assert d.flatten_dict == {"something": {}}

    d = ndict({"something": {}})
    assert d.dict == {"something": {}}
    assert d.flatten_dict == {"something": {}}
    d["something;node1"] = 1
    assert "something" not in d.flatten_dict


def test_bool():
    d = ndict()
    assert not d
    d["something"] = 1
    assert d
    del d["something"]
    assert not d


def test_paths():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))

    with open(TEST_ASSET / "init_paths.json", "r") as f:
        paths = json.load(f)
    assert d.paths == paths


def test_set_delimiter():
    with open(TEST_ASSET / "init.json", "r") as f:
        raw = json.load(f)
    with open(TEST_ASSET / "init_nested.json", "r") as f:
        nested = json.load(f)
    with open(TEST_ASSET / "delimiter_flatten.json", "r") as f:
        flatten = json.load(f)
    d = ndict(raw)
    d.delimiter = "."
    assert d.dict == nested, get_dict_compare_msg(d.dict, nested, indent=4, sort_keys=False)
    assert d.flatten_dict == flatten, get_dict_compare_msg(
        d.flatten_dict, flatten, indent=4, sort_keys=False
    )

    d1 = ndict(d)
    with open(TEST_ASSET / "init_flatten.json", "r") as f:
        flatten = json.load(f)
    assert d1.dict == nested, get_dict_compare_msg(d1.dict, nested, indent=4, sort_keys=False)
    assert d1.flatten_dict == flatten, get_dict_compare_msg(
        d1.flatten_dict, flatten, indent=4, sort_keys=False
    )


def test_states():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))
    d.delimiter = "+"
    states = d.state_dict()
    d1 = ndict().load_state_dict(states)
    with open(TEST_ASSET / "init_flatten.json", "r") as f:
        flatten = json.load(f)
    with open(TEST_ASSET / "init_nested.json", "r") as f:
        nested = json.load(f)
    assert d1.dict == nested, get_dict_compare_msg(d1.dict, nested, indent=4, sort_keys=False)
    assert d1.flatten_dict == flatten, get_dict_compare_msg(
        d1.flatten_dict, flatten, indent=4, sort_keys=False
    )


def test_get():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))
    with open(TEST_ASSET / "init_getitem.json", "r") as f:
        getitem_gt = json.load(f)

    for p, gt in getitem_gt.items():
        v = d.get(path=p)
        if isinstance(gt, dict):
            assert isinstance(v, ndict)
            v = v.dict
        assert v == gt

    for p, gt in getitem_gt.items():
        v = d.get(path=p, dict_as_ndict=False)
        if isinstance(gt, dict):
            assert isinstance(v, dict)
            v = v
        assert v == gt

    assert d.get("not_exist_path") is None
    assert d.get("not_exist_path", default=1) == 1
    assert d.get("not_exist_path;not_exist_path", default=1) == 1
    assert d.get("not_exist_path;not_exist_path") is None

    assert d.get(keys=["node1", "node2", ("node2", lambda tree, path: tree[path] + 1)]) == {
        "node1": "this",
        "node2": 2.0,
    }


def test_update_no_filting():
    d = ndict()
    with open(TEST_ASSET / "init_flatten.json", "r") as f:
        flatten = json.load(f)
    with open(TEST_ASSET / "init_nested.json", "r") as f:
        nested = json.load(f)
    d.update(flatten, ignore_missing=False, ignore_none=False)
    assert d.dict == nested, get_dict_compare_msg(d.dict, nested, indent=4, sort_keys=False)
    assert d.flatten_dict == flatten, get_dict_compare_msg(
        d.flatten_dict, flatten, indent=4, sort_keys=False
    )

    d = ndict()
    d.update(nested, ignore_missing=False, ignore_none=False)
    assert d.dict == nested, get_dict_compare_msg(d.dict, nested, indent=4, sort_keys=False)
    assert d.flatten_dict == flatten, get_dict_compare_msg(
        d.flatten_dict, flatten, indent=4, sort_keys=False
    )

    d1 = ndict(nested)
    d = ndict()
    d.update(d1, ignore_missing=False, ignore_none=False)
    assert d.dict == nested, get_dict_compare_msg(d.dict, nested, indent=4, sort_keys=False)
    assert d.flatten_dict == flatten, get_dict_compare_msg(
        d.flatten_dict, flatten, indent=4, sort_keys=False
    )

    d = ndict(nested)
    update_dict = {
        "node1": 1.0,
        "nested;node2": "something",
        "not_exist": "here",
        "node2": None,
        "node7": {"double": {"trible": {"leave": 345, "not_exist": 555}}},
    }
    d.update(update_dict, ignore_missing=False, ignore_none=False)
    with open(TEST_ASSET / "init_update1_gt_nested.json", "r") as f:
        nested_gt = json.load(f)
    with open(TEST_ASSET / "init_update1_gt_flatten.json", "r") as f:
        flatten_gt = json.load(f)
    assert d.dict == nested_gt
    assert d.flatten_dict == flatten_gt


def test_update_filting():
    d = ndict()
    with open(TEST_ASSET / "init_update_no_none_gt_nested.json", "r") as f:
        nested = json.load(f)
    with open(TEST_ASSET / "init_update_no_none_gt_flatten.json", "r") as f:
        flatten = json.load(f)
    d.update(flatten, ignore_missing=False, ignore_none=True)
    assert d.dict == nested, get_dict_compare_msg(d.dict, nested, indent=4, sort_keys=False)
    assert d.flatten_dict == flatten, get_dict_compare_msg(
        d.flatten_dict, flatten, indent=4, sort_keys=False
    )

    d = ndict()
    with open(TEST_ASSET / "init_flatten.json", "r") as f:
        flatten = json.load(f)
    d.update(flatten, ignore_missing=True, ignore_none=False)
    assert d.dict == {}
    assert d.flatten_dict == {}


def test_keys():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))

    with open(TEST_ASSET / "init_keys_depth1.json", "r") as f:
        gt = json.load(f)
    assert set(d.keys()) == set(gt)
    assert set(d.keys(1)) == set(gt)

    with open(TEST_ASSET / "init_keys_depth2.json", "r") as f:
        gt = json.load(f)
    assert set(d.keys(2)) == set(gt)

    with open(TEST_ASSET / "init_keys_depth3.json", "r") as f:
        gt = json.load(f)
    assert set(d.keys(3)) == set(gt)

    with open(TEST_ASSET / "init_keys_depth-1.json", "r") as f:
        gt = json.load(f)
    assert set(d.keys(-1)) == set(gt)
    assert set(d.keys(-1)) == set(d.flatten_dict.keys())


def test_values():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))
    with open(TEST_ASSET / "values_depth1.json", "r") as f:
        gt = json.load(f)
    assert d.values(1, dict_as_ndict=False) == gt
    assert all(v1 == v2 for v1, v2 in zip(d.values(1, dict_as_ndict=True), gt))
    assert all(v2 == v1 for v1, v2 in zip(d.values(1, dict_as_ndict=True), gt))
    with open(TEST_ASSET / "values_depth2.json", "r") as f:
        gt = json.load(f)
    assert d.values(2, dict_as_ndict=False) == gt
    assert all(v1 == v2 for v1, v2 in zip(d.values(2, dict_as_ndict=True), gt))
    assert all(v2 == v1 for v1, v2 in zip(d.values(2, dict_as_ndict=True), gt))
    with open(TEST_ASSET / "values_depth3.json", "r") as f:
        gt = json.load(f)
    assert d.values(3, dict_as_ndict=False) == gt
    assert all(v1 == v2 for v1, v2 in zip(d.values(3, dict_as_ndict=True), gt))
    assert all(v2 == v1 for v1, v2 in zip(d.values(3, dict_as_ndict=True), gt))
    with open(TEST_ASSET / "values_depth-1.json", "r") as f:
        gt = json.load(f)
    assert d.values(-1, dict_as_ndict=False) == gt
    assert all(v1 == v2 for v1, v2 in zip(d.values(-1, dict_as_ndict=True), gt))
    assert all(v2 == v1 for v1, v2 in zip(d.values(-1, dict_as_ndict=True), gt))
    for v, v_gt in zip(d.values(-1, dict_as_ndict=True), gt):
        if isinstance(v_gt, dict):
            assert isinstance(v_gt, ndict)
    for v, v_gt in zip(d.values(-1, dict_as_ndict=False), gt):
        if isinstance(v_gt, dict):
            assert isinstance(v_gt, dict)


def test_items():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))
    with open(TEST_ASSET / "init_keys_depth1.json", "r") as f:
        k_gt = json.load(f)
    with open(TEST_ASSET / "values_depth1.json", "r") as f:
        v_gt = json.load(f)
    for i, (k, v) in enumerate(d.items(1, dict_as_ndict=True)):
        assert k == k_gt[i]
        assert v == v_gt[i]
        if isinstance(v_gt[i], dict):
            assert isinstance(v, ndict)
            assert v.dict is d[k].dict
    for i, (k, v) in enumerate(d.items(1, dict_as_ndict=False)):
        assert k_gt[i] == k
        assert v_gt[i] == v
        if isinstance(v_gt[i], dict):
            assert isinstance(v, dict)
            assert v is d[k].dict

    with open(TEST_ASSET / "init_keys_depth2.json", "r") as f:
        k_gt = json.load(f)
    with open(TEST_ASSET / "values_depth2.json", "r") as f:
        v_gt = json.load(f)
    for i, (k, v) in enumerate(d.items(2)):
        assert k == k_gt[i]
        assert v == v_gt[i]
        if isinstance(v_gt[i], dict):
            assert isinstance(v, ndict)
            assert v.dict is d[k].dict
    for i, (k, v) in enumerate(d.items(2, dict_as_ndict=False)):
        assert k_gt[i] == k
        assert v_gt[i] == v
        if isinstance(v_gt[i], dict):
            assert isinstance(v, dict)
            assert v is d[k].dict

    with open(TEST_ASSET / "init_keys_depth3.json", "r") as f:
        k_gt = json.load(f)
    with open(TEST_ASSET / "values_depth3.json", "r") as f:
        v_gt = json.load(f)
    for i, (k, v) in enumerate(d.items(3)):
        assert k == k_gt[i]
        assert v == v_gt[i]
        if isinstance(v_gt[i], dict):
            assert isinstance(v, ndict)
            assert v.dict is d[k].dict
    for i, (k, v) in enumerate(d.items(3, dict_as_ndict=False)):
        assert k_gt[i] == k
        assert v_gt[i] == v
        if isinstance(v_gt[i], dict):
            assert isinstance(v, dict)
            assert v is d[k].dict

    with open(TEST_ASSET / "init_keys_depth-1.json", "r") as f:
        k_gt = json.load(f)
    with open(TEST_ASSET / "values_depth-1.json", "r") as f:
        v_gt = json.load(f)
    for i, (k, v) in enumerate(d.items(-1, dict_as_ndict=True)):
        assert k == k_gt[i]
        assert v == v_gt[i]
        if isinstance(v_gt[i], dict):
            assert isinstance(v, ndict)
            assert v.dict is d[k].dict
    for i, (k, v) in enumerate(d.items(-1, dict_as_ndict=False)):
        assert k_gt[i] == k
        assert v_gt[i] == v
        if isinstance(v_gt[i], dict):
            assert v is d[k].dict
            assert isinstance(v, dict)


def test_size():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))
    assert d.size(ignore_none=False) == len(d.keys())
    assert d.size(2, ignore_none=False) == len(d.keys(2))
    assert d.size(3, ignore_none=False) == len(d.keys(3))
    assert d.size(-1, ignore_none=False) == len(d.keys(-1))

    assert d.size(ignore_none=True) == len(d.keys()) - 1
    assert d.size(2, ignore_none=True) == len(d.keys(2)) - 2
    assert d.size(3, ignore_none=True) == len(d.keys(3)) - 4
    assert d.size(-1, ignore_none=True) == len(d.keys(-1)) - 4

    size = d.size(ignore_none=False)
    for k in d.keys():
        del d[k]
        assert d.size(ignore_none=False) == size - 1
        size = d.size(ignore_none=False)


def test_eq():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))
    d1 = deepcopy(d)
    assert d == d1

    flatten = {p: v for p, v in reversed(d.flatten_dict.items())}
    d1 = ndict(flatten)
    assert d == d1
    assert d == flatten

    d1["node1"] = "not exist"
    assert d != d1
    assert not d == d1

    d1 = ndict(flatten)
    d1["Not exist"] = "null"
    assert d != d1
    assert not d == d1


def test_diff():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))
    d1 = deepcopy(d)
    assert d.diff(d1) == {} and isinstance(d.diff(d1), dict)
    d1["node"] = "not"
    assert d.diff(d1) == {"node": (None, "not")}
    assert d1.diff(d) == {"node": ("not", None)}
    d1["node1"] = "not"
    assert d.diff(d1) == {"node": (None, "not"), "node1": ("this", "not")}
    assert d1.diff(d) == {"node": ("not", None), "node1": ("not", "this")}


def test_len():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))
    assert len(d) == d.size()

    for k in d.keys():
        del d[k]
        assert len(d) == d.size()


def test_print():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))
    assert str(d) == yaml.dump(d.dict, indent=2, sort_keys=False)
    assert d.json_str() == json.dumps(d.dict, sort_keys=False, indent=2)


def test_contain():
    with open(TEST_ASSET / "init.json", "r") as f:
        d = ndict(json.load(f))
    assert "node1" in d
    assert "not exist" not in d
    d = ndict()
    assert "node1" not in d
    assert "not exist" not in d


if __name__ == "__main__":
    # test_update_no_filting()
    test_setitem()



from __future__ import annotations

import json
from copy import deepcopy
from functools import reduce
from operator import getitem
from typing import Any, Callable, Optional, Union

import yaml

from .dict_traverse import traverse
from .stop_conditions import generate_depth_stop_condition


def in_or_callable(d: Union[ndict, dict], k: Union[str, Callable]) -> bool:
    return isinstance(k, Callable) or isinstance(k, str) and k in d


# TODO: depth change to max_depth.
# TODO: add comments.
# TODO: Change keys, values and items to generators.
class ndict:
    """Nested dictionary.

    Users shouldn't modify the underling data outside of ndict class. Value overwritten is enabled.

    Args:
        d (Optional[Union[ndict, dict]]): If d is a dict, do make sure the path separator is the givein delimiter if
            path is used as key.
        delimiter (str): Path separator. Can be any string.
    """

    ALL_MISSING_METHODS = ["ignore", "false", "exception"]

    def __init__(self, d: Optional[Union[ndict, dict]] = None, delimiter: str = ";") -> None:
        assert isinstance(delimiter, str), f"delimiter must be str, but recieved {type(delimiter)}"

        self._delimiter = delimiter
        self._d = {}
        self._flatten_dict = {}

        if isinstance(d, ndict):
            self.load_state_dict(d.state_dict())
        elif isinstance(d, dict):
            for k, v in d.items():
                self[k] = v
        else:
            assert d is None, f"Unexpected type {type(d)}"

    @property
    def dict(self) -> dict:
        """Underling dictionary. Do not change it!"""
        return self._d

    @property
    def flatten_dict(self) -> dict[str, Any]:
        """Flattened dictionary of {path: value} pairs."""
        return self._flatten_dict

    @property
    def paths(self) -> list[str]:
        """Get all possible paths."""

        def _path_action(tree: dict, res: list[str], node: Any, path: str, depth: int):
            if path is not None:
                res.append(path)

        res = []
        traverse(tree=self.dict, res=res, actions=_path_action)
        return res

    @property
    def delimiter(self) -> str:
        return self._delimiter

    @delimiter.setter
    def delimiter(self, delimiter: str) -> None:
        if delimiter == self._delimiter:
            return
        self._flatten_dict = {
            p.replace(self._delimiter, delimiter): v for p, v in self.flatten_dict.items()
        }
        self._delimiter = delimiter

    def state_dict(self) -> dict:
        return {"dict": self.dict, "flatten_dict": self.flatten_dict, "delimiter": self.delimiter}

    def load_state_dict(self, states: Union[dict, ndict]) -> Any:
        """The delimiter is only for properly initialize the object."""
        assert isinstance(
            states["delimiter"], str
        ), f"Unexpected delimiter type: {states['delimiter']}."
        delimiter = self.delimiter
        self._d = states["dict"]
        self._flatten_dict = states["flatten_dict"]
        self._delimiter = states["delimiter"]
        self.delimiter = delimiter
        return self

    def get(
        self,
        path: Optional[str] = None,
        default: Optional[Any] = None,
        keys: Optional[list[Union[str, tuple[str, Callable]]]] = None,
        dict_as_ndict: bool = True,
    ) -> Any:
        """Get values from the nested dictionary.

        Args:
            path (Optional[str] = None):
            keys (Optional[list[Union[str, tuple[str, Callable]]]] = None). Callables should accept self (ndict),
                path: str 2 arguments.
        """

        def _get_value_or_default(
            key: Union[str, Callable], default: Any, path: Optional[str] = None
        ) -> Any:
            if isinstance(key, str):
                try:
                    return self._get_node(key, dict_as_ndict=dict_as_ndict)
                except KeyError:
                    return default

            assert isinstance(key, Callable), f"Unexpected key type: {type(key)}."
            try:
                return key(self, path)
            except:
                return default

        assert (path is not None or keys is not None) and not (
            path is not None and keys is not None
        ), f"Users can only provide path or keys: path: {path is not None}, keys: {keys is not None}."

        if path is not None:
            try:
                return self._get_node(path, dict_as_ndict=dict_as_ndict)
            except KeyError:
                return default

        return {
            **{k: _get_value_or_default(k, default) for k in keys if isinstance(k, str)},  # type: ignore
            **{
                k[0]: _get_value_or_default(k[1], default, path=k[0])
                for k in keys  # type: ignore
                if isinstance(k, tuple)
            },
        }

    def update(
        self, d: Union[dict, ndict], ignore_none: bool = False, ignore_missing: bool = False
    ) -> None:
        """Could be slow at current stage.

        Note that if leaves of the d is Callable, the leaves will be invoked with self: ndict and path: str 2 arguments.
        """
        d = ndict(d).flatten_dict
        for p, v in d.items():
            if (v is None and ignore_none) or (p not in self.flatten_dict and ignore_missing):
                continue
            self[p] = v(self, p) if isinstance(v, Callable) else v

    def keys(self, max_depth: int = 1) -> list[str]:
        """Return a list of leave and depth <= depth"""

        def _keys_action(tree: dict, res: list[str], node: Any, path: str, depth: int):
            if path is not None and (not isinstance(node, dict) or depth == max_depth):
                res.append(path)

        res = []
        traverse(tree=self.dict, res=res, actions=_keys_action, depth=max_depth)
        return res

    def values(self, max_depth: int = 1, dict_as_ndict: bool = True) -> list[Any]:
        def _values_action(tree: dict, res: list[Any], node: Any, path: str, depth: int):
            if path is not None and (not isinstance(node, dict) or depth == max_depth):
                if dict_as_ndict and isinstance(node, dict):
                    node = ndict(delimiter=self.delimiter).load_state_dict(
                        {
                            "dict": node,
                            "flatten_dict": self._get_flatten_dict_of_subtree(path),
                            "delimiter": self.delimiter,
                        }
                    )
                res.append(ndict(node) if dict_as_ndict and isinstance(node, dict) else node)

        res = []
        traverse(tree=self.dict, res=res, actions=_values_action, depth=max_depth)
        return res

    def items(self, max_depth: int = 1, dict_as_ndict: bool = True) -> list[tuple[str, Any]]:
        return list(
            zip(
                self.keys(max_depth=max_depth),
                self.values(max_depth=max_depth, dict_as_ndict=dict_as_ndict),
            )
        )

    # May let the users to cumstomize the conditions.
    def size(self, max_depth: int = 1, ignore_none: bool = False) -> int:
        def _size_action(tree: dict, res: list[int], node: Any, path: str, depth: int):
            if (
                path is not None
                and (not isinstance(node, dict) or depth == max_depth)
                and (not ignore_none or ignore_none and node is not None)
            ):
                res[0] += 1

        res = [0]
        traverse(tree=self.dict, res=res, actions=_size_action, depth=max_depth)
        return res[0]

    def diff(self, d: Union[ndict, dict]) -> dict[str, tuple[Any, Any]]:
        """Compare the leaves."""
        d = ndict(d)
        res = {}
        for p, v1 in self.flatten_dict.items():
            if p not in d.flatten_dict:
                res[p] = (v1, None)
            elif v1 != d[p]:
                res[p] = (v1, d[p])
        res.update({p: (None, v) for p, v in d.flatten_dict.items() if p not in self})
        return res

    def json_str(self, sort_keys: bool = False, indent: int = 2) -> str:
        return json.dumps(self.dict, sort_keys=sort_keys, indent=indent)

    def __getitem__(self, key: Union[str, int]) -> Any:
        path = key if isinstance(key, str) else list(self.keys())[key]
        return self._get_node(path, dict_as_ndict=True)

    def __delitem__(self, path: str) -> None:
        path_list = path.split(self._delimiter)
        d = self._get_node(path_list[:-1], dict_as_ndict=False)
        del d[path_list[-1]]
        if path:
            self._flatten_dict = {
                p: v for p, v in self._flatten_dict.items() if not p.startswith(path)
            }
        else:
            self._flatten_dict = {
                p: v for p, v in self._flatten_dict.items() if not p.startswith(";") and p != ""
            }
        if len(d) == 0:
            if len(path_list) > 1:
                parent_path = self._delimiter.join(path_list[:-1])
                self._flatten_dict[parent_path] = {}

    # TODO A more efficient approach.
    def __setitem__(self, path: str, value: Any) -> None:
        """Update values of the corresponding path.

        Args:
            path (str): Path can be an existed or non-exist path. If it's existed path and the corresponding values is
                not a dictionary, then the original value will be overwrittern.
            value (Any): The value for that path.
        """
        assert isinstance(path, str), f"Path can only be str, recieved {type(path)}."
        path_list = path.split(self.delimiter)

        # Adjust dict.
        d = self._d
        for i, node in enumerate(path_list[:-1]):
            if node not in d:
                d[node] = {}
            elif not isinstance(d[node], dict):
                d[node] = {}
                tmp_p = self.delimiter.join(path_list[: i + 1])
                del self.flatten_dict[tmp_p]
            d = d[node]

        to_be_delete_node = [
            f"{path}{self.delimiter}{k}" for k in self._get_flatten_dict_of_subtree(path).keys()
        ]
        for k in to_be_delete_node:
            del self._flatten_dict[k]
        parent_path = self.delimiter.join(path_list[:-1])
        if parent_path in self._flatten_dict:
            assert (
                isinstance(self._flatten_dict[parent_path], dict)
                and not self._flatten_dict[parent_path]
            )
            del self._flatten_dict[parent_path]

        # Adjust flatten dict.
        if isinstance(value, Union[dict, ndict]):
            tmp = ndict(value)
            d[path_list[-1]] = tmp.dict
            if tmp.flatten_dict:
                for p, v in tmp.flatten_dict.items():
                    combined_path = self._delimiter.join([path, p])
                    self._flatten_dict[combined_path] = v
            else:
                self._flatten_dict[path] = {}
        else:
            d[path_list[-1]] = value
            self._flatten_dict[path] = value

    def __len__(self) -> int:
        return len(self.dict)

    def __bool__(self) -> bool:
        return bool(len(self) > 0)

    def __contains__(self, path: str) -> bool:
        if path in self.flatten_dict:
            return True

        nodes = path.split(self._delimiter)
        d = self.dict
        for n in nodes:
            if n not in d:
                return False
            d = d[n]
        return True

    def __eq__(self, other: Union[dict, ndict]) -> bool:
        other = ndict(other)
        return self.flatten_dict == other.flatten_dict

    def __str__(self) -> str:
        return yaml.dump(self.dict, sort_keys=False, indent=2)

    def __repr__(self) -> str:
        return f"<Nested dictionary of {len(self)} leaves.>: {self.dict}"

    def _get_node(self, path: Union[str, list[str]], dict_as_ndict: bool) -> Any:
        """Return the value of a particular path.

        Return
            Node value. If the node is a dictionary, __class__(node) will be returned.
        """
        if isinstance(path, list):
            path_list = path
            path = self.delimiter.join(path)
        elif isinstance(path, str):
            path_list = path.split(self.delimiter)
        else:
            raise ValueError(f"Unexpected path type: {type(path)}")

        v = reduce(getitem, path_list, self._d)

        if isinstance(v, dict) and dict_as_ndict:
            states = {
                "dict": v,
                "flatten_dict": self._get_flatten_dict_of_subtree(path),
                "delimiter": self.delimiter,
            }
            return ndict(delimiter=self.delimiter).load_state_dict(states)
        else:
            return v

    def _get_flatten_dict_of_subtree(self, prefix: str) -> dict[str, Any]:
        prefix = f"{prefix}{self.delimiter}"
        return deepcopy(
            {
                path[len(prefix) :]: value
                for path, value in self.flatten_dict.items()
                if path.startswith(prefix) and len(path) >= len(prefix)
            }
        )

    def _flatten(self) -> dict[str, Any]:
        def flatten_action(tree: dict, res: dict, node: Any, path: str, depth: int) -> None:
            if not isinstance(node, dict):
                res[path] = node

        res = {}
        traverse(self.dict, res, flatten_action)
        return res



from typing import Any, Union, Callable, Optional
from .stop_conditions import generate_depth_stop_condition


### Algs ###
def dfs(
    tree: dict,
    res: Any,
    node: Any,
    path: str,
    depth: int,
    action: Callable,
    stop_condition: Callable,
) -> Any:
    """Depth-first traverse.

    Args:
        tree (Any): The whole dictionary. This argument is used to provide global information and
            modify the tree.
        res (Any): Current results.
        node (Any): Current node (value).
        path (str): Current path (key)
        depth (int): Current depth (0 at root)
        action (callable): Action.
        stop_condition (callable): Whether should stop the traverse process.

    Returns:
        Any: Results
    """
    if stop_condition(res, node, path, depth):
        return res

    action(tree, res, node, path, depth)

    if isinstance(node, dict):
        for k, v in node.items():
            next_path = k if path is None else ";".join([path, k])
            dfs(tree, res, v, next_path, depth + 1, action, stop_condition)
    else:
        return res


def traverse(
    tree: dict,
    res: Any,
    actions: Union[Callable, tuple[Callable, dict[str, Callable]]],
    depth: int = -1,
    stop_condition: Optional[Union[list[Callable], Callable]] = None,
    alg: Callable = dfs,
) -> Any:
    """Go through nodes in the tree and apply actions / collect data.

    Args:
        res (Any): Where the final results will be.
        actions (Union[callable, tuple[callable, dict[str, callable]]]): action applied on all nodes or A tuple of
            (default action, {path: path actions}). Each callable should accept: tree, res, node, path, depth 5
            arguments. If the dictionary is modified, the action function must also update the flatten dictionary at
            the same time.
        depth (Optional[int], optional): Maximum traverse depth. Defaults to -1, which means traverse all depth.
        stop_condition (Optional[Union[list[callable], callable]], optional): Callable which returns bool to
            determine whether the tranverse should be stopped. This callable should accept res, node, path, depth 4
            arguments. It can also be a list of callables. In that case all the callable must return True to stop
            the traverse process. Defaults to None which means no extra conditions other depth.
        alg (callable, optional): Algorithm used to traverse the tree. Currently only support dfs. It should accept
            res, node, path, depth, actions and stop_condition 6 arguments. Defaults to dfs.

    Returns:
        Any: The results
    """
    action = _generate_action_pipeline(actions)
    stop_condition = _generate_stop_condition_pipeline(stop_condition, depth)

    return alg(tree, res, tree, None, 0, action, stop_condition)


def _generate_action_pipeline(
    actions: Union[Callable, tuple[Callable, dict[str, Callable]]]
) -> Callable:
    if isinstance(actions, tuple):
        default_action = actions[0]
        if len(actions) == 2:
            specific_actions = actions[1]
    else:
        default_action = actions
        specific_actions = {}

    def action_pipeline(tree: Any, res: Any, node: Any, path: str, depth: int) -> None:
        if path in specific_actions:
            specific_actions[path](tree, res, node, path, depth)
        else:
            default_action(tree, res, node, path, depth)

    return action_pipeline


def _generate_stop_condition_pipeline(
    conditions: Optional[Union[list[Callable], Callable]], max_depth: int = -1
) -> Callable:
    conditions = [conditions] if isinstance(conditions, Callable) else conditions
    conditions = [] if conditions is None else conditions
    conditions.append(generate_depth_stop_condition(max_depth))

    def stop_condition_pipeline(res: Any, node: Any, path: str, depth: int) -> bool:
        for cond in conditions:
            if cond(res, node, path, depth):
                return True
        return False

    return stop_condition_pipeline



from typing import Any, Callable


def generate_depth_stop_condition(max_depth: int = -1) -> Callable:
    def depth_stop_condition(res: Any, node: Any, path: str, depth: int) -> bool:
        return max_depth != -1 and depth > max_depth

    return depth_stop_condition


def generate_subtree_stop_condition(subtree_path: str) -> Callable:
    def subtree_stop_condition(res: Any, node: Any, path: str, depth: int) -> bool:
        return path.startswith(subtree_path)

    return subtree_stop_condition
